#!/usr/bin/env python

import math
import string
import os
import sys
from optparse import OptionParser
import util
import asa
from vector3d import Vector3d, pos_distance, dot
from molecule import Molecule, Atom, add_radii, get_center, get_width


class Grid:

  def __init__(self, grid_spacing, width, center):
    self.center = center
    self.width = width
    self.grid_spacing = grid_spacing
    self.inv_grid_spacing = 1.0 / self.grid_spacing
    self.n_point = int(math.ceil(self.width * self.inv_grid_spacing))
    self.half_n_point = self.n_point // 2

    self.exclude_array = []
    for i in range(self.n_point):
      self.exclude_array.append([])
      for j in range(self.n_point):
        self.exclude_array[-1].append([])
        for k in range(self.n_point):
          self.exclude_array[-1][-1].append([False, False])
      
  def indices(self, pos):
    return ((pos.x-self.center.x)*self.inv_grid_spacing + self.half_n_point,
            (pos.y-self.center.y)*self.inv_grid_spacing + self.half_n_point,
            (pos.z-self.center.z)*self.inv_grid_spacing + self.half_n_point)

  def pos(self, i, j, k):
    return Vector3d(
            self.center.x + (i-self.half_n_point)*self.grid_spacing,
            self.center.y + (j-self.half_n_point)*self.grid_spacing,
            self.center.z + (k-self.half_n_point)*self.grid_spacing)

  def is_grid_point_near_sphere(self, i, j, k, pos, r):
    d_x = self.center.x + (i-self.half_n_point)*self.grid_spacing - pos.x
    d_y = self.center.y + (j-self.half_n_point)*self.grid_spacing - pos.y 
    d_z = self.center.z + (k-self.half_n_point)*self.grid_spacing - pos.z
    return (d_x*d_x + d_y*d_y + d_z*d_z) < r*r
    
  def int_range(self, low_f, high_f):
    low = max(0, math.floor(low_f))
    high = min(self.n_point, math.ceil(high_f) + 1)
    return range(int(low), int(high))

  def exclude_sphere(self, pos, r):
    low = Vector3d(pos.x - r, pos.y - r, pos.z - r)
    low_i, low_j, low_k = self.indices(low)
    high = Vector3d(pos.x + r, pos.y + r, pos.z + r)
    high_i, high_j, high_k = self.indices(high)
    for i in self.int_range(low_i, high_i):
      for j in self.int_range(low_j, high_j):
        for k in self.int_range(low_k, high_k):
          if not self.exclude_array[i][j][k][0]:
            if self.is_grid_point_near_sphere(i, j, k, pos, r):
              self.exclude_array[i][j][k][0] = True

  def permutation(self, i, j, k, dim):
    if dim == 0:
      return i, j, k
    if dim == 1:
      return j, k, i
    if dim == 2:
      return k, i, j

  def drill_in_dim(self, is_reversed, i, j, dim):
    drill_range = range(self.n_point)
    if is_reversed:
      drill_range.reverse()
    for k in drill_range:
      a, b, c = self.permutation(i, j, k, dim)
      if self.exclude_array[a][b][c][0]:
        return
      self.exclude_array[a][b][c][1] = True

  def exclude_edge_to_interior(self):
    for i in range(self.n_point):
      for j in range(self.n_point):
        self.drill_in_dim(True,  i, j, 0)
        self.drill_in_dim(False, i, j, 0)
        self.drill_in_dim(True,  i, j, 1)
        self.drill_in_dim(False, i, j, 1)
        self.drill_in_dim(True,  i, j, 2)
        self.drill_in_dim(False, i, j, 2)
          
  def is_excluded(self, i, j, k):
    exclusion = self.exclude_array[i][j][k]
    return exclusion[0] or exclusion[1]
    
  def is_surrounded(self, i, j, k):
    indices_list = [(i,j,k), 
                    (i+1,j,k), (i-1,j,k), 
                    (i,j+1,k), (i,j-1,k), 
                    (i,j,k-1), (i,j,k+1)]
    in_grid = lambda x: x >= 0 and x < self.n_point
    for (a, b, c) in indices_list:
      if in_grid(a) and in_grid(b) and in_grid(c):
        if self.is_excluded(a, b, c):
          return False
    return True

  def exclude_surrounded(self, skip):
    surrounded_grid_points = []
    for i in range(self.n_point):
      for j in range(self.n_point):
        for k in range(self.n_point):
          if self.is_surrounded(i,j,k):
            surrounded_grid_points.append([i,j,k])
    for (i,j,k) in surrounded_grid_points:
      if i % skip == 0 and j % skip == 0 and k % skip == 0:
        continue
      self.exclude_array[i][j][k][0] = True

  def exclude_points_in_constraint(self, constraint_fn):
    for i in range(self.n_point):
      for j in range(self.n_point):
        for k in range(self.n_point):
          if not grid.exclude_array[i][j][k][0]:
            if not constraint_fn(grid.pos(i,j,k)):
              grid.exclude_array[i][j][k][0] = True

  def make_mol(self, res_type, atom_type):
    mol = Molecule()
    i_res = 1
    for i in range(self.n_point):
      for j in range(self.n_point):
        for k in range(self.n_point):
          if not (self.is_excluded(i,j,k)):
            hollow_atom = Atom()
            hollow_atom.pos = grid.pos(i,j,k)
            hollow_atom.type = atom_type
            hollow_atom.element = ""
            for c in atom_type[:2]:
              if not c.isdigit() and c != " ":
                hollow_atom.element += c
            hollow_atom.res_type = res_type
            hollow_atom.res_num = i_res
            hollow_atom.num = i_res
            mol.insert_atom(hollow_atom)
            i_res += 1
    return mol


def exclude_atoms_from_grid(grid, atoms, probe):
  for atom in atoms:
    grid.exclude_sphere(atom.pos, atom.radius + probe)
  

def exclude_surface(grid, atoms, probe):
  test_point = Vector3d()
  sphere_points = asa.generate_sphere_points(960)
  for i, atom_i in enumerate(atoms):
    is_interior_atom = "asa" in atom_i.__dict__ and atom_i.asa < 9
    if not is_interior_atom:
      neighbor_indices = asa.find_neighbor_indices(atoms, probe, i)
      n_neighbor = len(neighbor_indices)
      j_closest_neighbor = 0
      radius = probe + atom_i.radius
      for point in sphere_points:
        is_point_accessible = True
        test_point.x = point[0]*radius + atom_i.pos.x
        test_point.y = point[1]*radius + atom_i.pos.y
        test_point.z = point[2]*radius + atom_i.pos.z
        cycled_indices = range(j_closest_neighbor, n_neighbor)
        cycled_indices.extend(range(j_closest_neighbor))
        for j in cycled_indices:
            atom_j = atoms[neighbor_indices[j]]
            r = atom_j.radius + probe
            d_x = test_point.x - atom_j.pos.x
            d_y = test_point.y - atom_j.pos.y
            d_z = test_point.z - atom_j.pos.z
            if d_x*d_x + d_y*d_y + d_z*d_z < r*r:
                j_closest_neighbor = j
                is_point_accessible = False
                break
        if is_point_accessible:
            grid.exclude_sphere(test_point, probe)

  
def calculate_average_bfactor(grid_chain, protein_atoms, bfactor_probe):
  max_bfactor = 0.0
  for atom in protein_atoms:
    if atom.bfactor > max_bfactor:
      max_bfactor = atom.bfactor
  for grid_atom in grid_chain.atoms():
      bfactors = []
      for protein_atom in protein_atoms:
        if protein_atom.element != "H":
          radius = bfactor_probe
          dist = pos_distance(protein_atom.pos, grid_atom.pos)
          if dist < radius:
            bfactors.append(protein_atom.bfactor)
      n_bfactor = len(bfactors)
      if n_bfactor == 0:
        grid_atom.bfactor = max_bfactor
      else:              
        grid_atom.bfactor = sum(bfactors)/float(n_bfactor)


def add_asa(atoms, probe):
  areas = asa.calculate_asa(atoms, probe)
  for area, atom in zip(areas, atoms):
    atom.asa = area
    

def get_molecule(pdb):
  txt = open(pdb, 'r').read()
  txt = util.strip_other_nmr_models(txt)
  txt = util.strip_lines(txt, lambda l: l.startswith('ANISOU'))
  txt = util.strip_lines(txt, lambda l: l.startswith('CONECT'))
  txt = util.strip_lines(txt, lambda l: l.startswith('MASTER'))
  txt = util.strip_alternative_atoms(txt)
  bare_fname = util.temp_fname('.pdb')
  open(bare_fname, 'w').write(txt)
  mol = Molecule(bare_fname)
  if os.path.isfile(bare_fname):
    os.remove(bare_fname)
  return mol


def find_atom(atoms, chain, res_num, atom_type):
  if chain == "":
    chain = " "
  for atom in atoms:
    if atom.chain_id == chain \
      and atom.res_num == res_num \
      and atom.type == atom_type:
      return atom
  raise "Can't find '%s' atom %s of res %d" % (chain, atom_type, res_num)


def get_sphere_constraint_fn(center, radius):
  return lambda pos: pos_distance(center, pos) <= radius
  
  
def get_cylinder_constraint_fn(center1, center2, radius):
  axis12 = center2 - center1

  def cylinder_constraint_fn(pos):
    pos1 = pos - center1
    if dot(pos1, axis12) < 0:
      return False
    pos1_perp = pos1.perpendicular_vec(axis12)
    if pos1_perp.length() > radius:
      return False
    pos2 = pos - center2  
    if dot(pos2, axis12) > 0:
      return False
    return True
  
  return cylinder_constraint_fn

  
# load default values from "hollow.txt"
module_dir = os.path.dirname(__file__)
default_fname = os.path.join(module_dir, "hollow.txt")
defaults = util.read_parameters(default_fname)

copyright = """
Hollow (c) 2007 Bosco Ho & Franz Gruswitz.

Generates a PDB of fake atoms that fill voids, pockets,
clefts and channels of a protein structure.
"""

parser = OptionParser(usage="usage: %prog [options] pdb")
parser.add_option("-g", dest="grid_spacing", type="float", 
                  help="grid spacing in angstroms" 
                       " (default %.2f angstroms;"
                       " suggested 0.2 for final resolution)" % \
                       defaults.grid_spacing,
                  default=defaults.grid_spacing)
parser.add_option("-c", dest="constraint_file",
                  help="read CONSTRAINT_FILE for grid constraints",
                  default='')
parser.add_option("-o", dest="out_pdb",
                  help="output hollow spheres to OUT_PDB" ,
                  default='')
parser.add_option("-p", dest="interior_probe", type="float",
                  help="radius of ball to explore cavities"
                       " (default %.2f angstroms = 95%% x radius"
                       " of output atom type suggested)" % \
                       defaults.interior_probe,
                  default=defaults.interior_probe)
parser.add_option("-s", dest="surface_probe", type="float",
                  help="radius of probe to roll over surface"
                       " used to define depressions"
                       " (default %.2f angstroms)" % \
                       defaults.surface_probe,
                  default=defaults.surface_probe)
parser.add_option("-w", dest="process_waters",
                  action="store_true", 
                  help="process water molecules"
                       " (no-flag=remove waters from surface calculation;"
                       " flag=include waters in protein surface)",
                  default=False)
parser.add_option("-b", dest="bfactor_probe", type="float",
                  help="radius around a grid point, in which the"
                       " b-factors of heavy atoms are averaged"
                       " (default=0.0=off; suggested=4.0)",
                  default=0.0)
options, args = parser.parse_args()                
if len(args) == 0:
  print copyright
  parser.print_help()
  print
  sys.exit(1)
timer = util.Timer()


# load protein and get protein parameters
pdb = args[0]
base = pdb.replace('.pdb', '')
print "Loading " + pdb
mol = get_molecule(pdb)
atoms = mol.atoms()
add_radii(atoms)


# Setup constraints and grid size
timer.start()
if options.constraint_file:
  print "Loading constraints from %s", options.constraint_file
  constraints = util.read_parameters(options.constraint_file)
  if constraints.type == 'sphere':
    atom1 = find_atom(atoms, constraints.chain1, 
                      constraints.res_num1, 
                      constraints.atom1)
    radius = constraints.radius
    constraint_fn = get_sphere_constraint_fn(atom1.pos, radius)
    center = atom1.pos
    width = 2.0*constraints.radius + 2.0*options.grid_spacing
    radius = radius - 3.0*options.grid_spacing
    inner_constraint_fn = get_sphere_constraint_fn(atom1.pos, radius)
  elif constraints.type == 'cylinder':
    atom1 = find_atom(atoms, constraints.chain1, 
                      constraints.res_num1, 
                      constraints.atom1)
    atom2 = find_atom(atoms, constraints.chain2, 
                      constraints.res_num2, 
                      constraints.atom2)

    axis12 = atom2.pos - atom1.pos

    offset1 = -axis12.normal_vec()
    offset1.scale(constraints.axis_offset1)
    center1 = atom1.pos + offset1

    offset2 = axis12.normal_vec()
    offset2.scale(constraints.axis_offset2)
    center2 = atom2.pos + offset2

    center = center1 + center2
    center.scale(0.5)
    radius = constraints.radius
    constraint_fn = get_cylinder_constraint_fn(center1, 
                                       center2, radius)

    half_length = pos_distance(center, center1)
    width = 2.0*math.sqrt(half_length*half_length +
                  constraints.radius*constraints.radius) \
            + 2.0*options.grid_spacing
            
    border_length = 3.0*options.grid_spacing
    center1 = center1 + axis12.normal_vec().scaled_vec(border_length)
    center2 = center2 - axis12.normal_vec().scaled_vec(border_length)
    radius = radius - border_length
    inner_constraint_fn = get_cylinder_constraint_fn(center1,
                                      center2, radius)
  else:
    raise "Don't understand constraint type"
else: # no constraints                    
  center = get_center(atoms)
  width = get_width(atoms, center)
n_point = width / options.grid_spacing
print "Setting up grid: %d x %d x %d, spacing %.3f, width %.1f" \
       % (n_point, n_point, n_point, options.grid_spacing, width)
grid = Grid(options.grid_spacing, width, center)
print "   >>>>>", timer.str()


# Exclude protein from grid
if not options.process_waters:
  print "Skipping water molecules"
  atoms = [a for a in atoms if a.res_type != "HOH"]
print "Excluding protein bulk from grid with %.1f angstrom probe" \
        % options.interior_probe
timer.start()
exclude_atoms_from_grid(grid, atoms, options.interior_probe)
print "   >>>>>", timer.str()
  
  
if options.constraint_file:
  # Eliminate all grid points outside constraints
  print "Excluding grid points outside constraint"
  timer.start()
  grid.exclude_points_in_constraint(constraint_fn)
  print "   >>>>>", timer.str()


is_calculate_asa_shell = True
if options.constraint_file:
  if not constraints.remove_asa_shell:
    is_calculate_asa_shell = False
if is_calculate_asa_shell:
  # Roll large ball over surface residues to eliminate
  # grid points over surface, then drill in from the
  # edge to eliminate the rest
  print "Calculating asa of atoms in protein"
  timer.start()
  add_asa(atoms, 1.4)
  print "   >>>>>", timer.str()
  timer.start()
  print "Excluding surface shell from grid with %.1f angstrom probe" \
          % options.surface_probe
  exclude_surface(grid, atoms, options.surface_probe)
  print "   >>>>>", timer.str()
  print "Excluding edges"
  grid.exclude_edge_to_interior()


print "Excluding surrounded points"
timer.start()
hole_size = int(2.0*1.4 / options.grid_spacing)
grid.exclude_surrounded(hole_size)
print "   >>>>>", timer.str()


# Make hollow spheres from grid-points
grid_chain = grid.make_mol(defaults.res_type, 
                           defaults.atom_type)

if options.bfactor_probe:
  timer.start()
  print "Averaging b-factors for each grid point"
  if options.constraint_file:
    atoms = [a for a in atoms if constraint_fn(a.pos)]
  calculate_average_bfactor(grid_chain, atoms, 
                      options.bfactor_probe)  
  print "   >>>>>", timer.str()

if options.constraint_file:
    for atom in grid_chain.atoms():
      if inner_constraint_fn(atom.pos):
        atom.occupancy = 1.0
      else:
        atom.occupancy = 0.0

if not options.out_pdb:
  options.out_pdb = base + '-hollow.pdb'

print "Saving hollow spheres to", options.out_pdb
f = open(options.out_pdb, 'w')
i_chain = 0
for i_atom, atom in enumerate(grid_chain.atoms()):
  atom.num = i_atom + 1
  atom.res_num = atom.num % 10000
  if atom.num > 1 and atom.num % 10000 == 0:
    i_chain += 1
  atom.chain_id = string.ascii_uppercase[i_chain]
  if defaults.atom_field.startswith("ATOM"):
    atom.is_hetatm = False
  else:
    atom.is_hetatm = True
  f.write(atom.pdb_str() + '\n')
f.close()
