#!/usr/bin/env python

import math, string
from vector3d import Vector3d, pos_distance, dot
from molecule import Molecule, Atom, add_radii, get_center, get_width


class Grid:

  def __init__(self, grid_spacing, width, center):
    self.width = width
    self.half_width = self.width / 2
    self.set_center(center)
    self.grid_spacing = grid_spacing
    self.inv_grid_spacing = 1.0 / self.grid_spacing
    self.n_point = int(math.ceil(self.width * self.inv_grid_spacing))

    self.exclude_array = []
    for i in range(self.n_point):
      self.exclude_array.append([])
      for j in range(self.n_point):
        self.exclude_array[-1].append([])
        for k in range(self.n_point):
          self.exclude_array[-1][-1].append(False)
      
  def reset(self):
    for i in range(self.n_point):
      for j in range(self.n_point):
        for k in range(self.n_point):
          self.exclude_array[i][j][k] = False
    
  def set_center(self, center):
    self.low = center.copy()
    self.low.x -= self.half_width
    self.low.y -= self.half_width
    self.low.z -= self.half_width
    
  def indices(self, pos):
    return ((pos.x-self.low.x)*self.inv_grid_spacing - 0.5,
            (pos.y-self.low.y)*self.inv_grid_spacing - 0.5,
            (pos.z-self.low.z)*self.inv_grid_spacing - 0.5)

  def pos(self, i, j, k):
    return Vector3d(
            self.low.x + (i+0.5)*self.grid_spacing,
            self.low.y + (j+0.5)*self.grid_spacing,
            self.low.z + (k+0.5)*self.grid_spacing)

  def is_grid_point_near_sphere(self, i, j, k, pos, r):
    d_x = self.low.x + (i+0.5)*self.grid_spacing - pos.x
    d_y = self.low.y + (j+0.5)*self.grid_spacing - pos.y 
    d_z = self.low.z + (k+0.5)*self.grid_spacing - pos.z
    return (d_x*d_x + d_y*d_y + d_z*d_z) < r*r
    
  def int_range(self, low_f, high_f):
    low = max(0, math.floor(low_f))
    high = min(self.n_point, math.ceil(high_f) + 1)
    return range(int(low), int(high))

  def exclude_sphere(self, pos, r):
    low = Vector3d(pos.x - r, pos.y - r, pos.z - r)
    low_i, low_j, low_k = self.indices(low)
    high = Vector3d(pos.x + r, pos.y + r, pos.z + r)
    high_i, high_j, high_k = self.indices(high)
    for i in self.int_range(low_i, high_i):
      for j in self.int_range(low_j, high_j):
        for k in self.int_range(low_k, high_k):
          if not self.exclude_array[i][j][k]:
            if self.is_grid_point_near_sphere(i, j, k, pos, r):
              self.exclude_array[i][j][k] = True

  def n_excluded(self):
    result = 0
    for i in range(self.n_point):
      for j in range(self.n_point):
        for k in range(self.n_point):
          if self.exclude_array[i][j][k]:
            result += 1
    return result
    
        
def volume(atoms, grid_spacing):
  center = get_center(atoms)
  width = get_width(atoms, center) + 3.0
  half_width = width / 2
  low = center - Vector3d(half_width, half_width, half_width)

  block_n_point = 40
  block_width = grid_spacing*block_n_point
  n_block = int(width / grid_spacing / block_n_point) + 2

  grid = Grid(grid_spacing, block_width, center)
  block_center = Vector3d()

  total_volume = 0.0

  for i in range(n_block):
    for j in range(n_block):
      for k in range(n_block):
        block_center.x = low.x + (i+0.5)*block_width
        block_center.y = low.y + (j+0.5)*block_width
        block_center.z = low.z + (k+0.5)*block_width
        grid.set_center(block_center)
        grid.reset()
        for atom in atoms:
          grid.exclude_sphere(atom.pos, atom.radius)
        fraction = grid.n_excluded()/float(block_n_point**3)
        max_block_volume = block_width*block_width*block_width
        block_volume = max_block_volume * fraction
        total_volume += block_volume

  return total_volume


if __name__ == "__main__":
  import sys
  import molecule

  usage = \
  """
  Calculates the volume of the atoms in a PDB file.
  Copyright (c) 2007 Bosco Ho.  

  Usage: volume.py in_pdb [spacing]
  
  -spacing    the spacing in the grid used to calculate the 
              volume grid (default=0.5). The smaller value,
              the more accurate, but slower.
              
  """

  if len(sys.argv) < 2:
    print usage
    sys.exit(1)
    
  mol = molecule.Molecule(sys.argv[1])
  atoms = mol.atoms()
  molecule.add_radii(atoms)
  if len(sys.argv) > 2:
    spacing = float(sys.argv[2])
  else:
    spacing = 0.5

  print "%.1f angstroms cubed" % volume(atoms, spacing)
